# Copyright 2022 Animatea
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Dot_tense service converters."""
from __future__ import annotations

__all__ = [
    "AbstractParticleConverter",
    "DigitValueConverter",
    "BooleanValueConverter",
    "ListValueConverter",
    "AbstractParticleValueConverter",
    "PARTICLE_CONVERTERS",
    "GETATTRIBUTE_VALUE_CONVERTERS",
    "VCONVERTER_CONSTS",
]

import abc
from typing import Any, Hashable

from aiotense.service_layer.safe_eval import SafelyExpEvalute

VCONVERTER_CONSTS: dict[str, int] = {
    "second": 1,
    "minute": 60,
    "hour": 60 * 60,
    "day": 60 * 60 * 24,
    "week": 60 * 60 * 24 * 7,
    "year": 60 * 60 * 24 * 365,
}


class AbstractParticleConverter(abc.ABC):
    """Base converter for particles (tokens)."""

    @abc.abstractmethod
    def convert(self, value: Any) -> Any:
        ...


class AbstractParticleValueConverter(abc.ABC):
    """Base converter for particle values.

    We can say that it is a subconverter.
    For example, if AbstractParticleConverter
    will convert the expression `variable = value`,
                                            ^^^^^
    then AbstractParticleValueConverter will convert
    the value of `variable` -- `value`.
                                ^^^^^
    """

    @abc.abstractmethod
    def matches(self, value: str) -> bool:
        ...

    @abc.abstractmethod
    def convert(self, value: str) -> Any:
        ...


class GetattributeParticleConverter(AbstractParticleConverter):
    @staticmethod
    def _parse_exp(value: str) -> tuple[str, str]:
        key, value = [_.strip() for _ in value.split("=")]
        return key, value

    def convert(self, value: str) -> dict[Hashable, Any]:
        key, value = self._parse_exp(value)
        for vconverter in GETATTRIBUTE_VALUE_CONVERTERS:
            if vconverter.matches(value):
                return {key: vconverter.convert(value)}

        raise ValueError(f"Can't find converter for {value!r}")


class DigitValueConverter(AbstractParticleValueConverter):
    def matches(self, value: str) -> bool:
        return value.isdigit()

    def convert(self, value: str) -> int:
        return int(value)


class BooleanValueConverter(AbstractParticleValueConverter):
    def matches(self, value: str) -> bool:
        return value.lower() in {"true", "false"}

    def convert(self, value: str) -> bool:
        return value.lower() == "true"


class ListValueConverter(AbstractParticleValueConverter):
    def matches(self, value: str) -> bool:
        return len(value.split(",")) > 1

    def convert(self, value: str) -> list[str] | str:
        by_comma = value.split(",")
        return [i.strip() for i in by_comma if i and not i.isspace()]


class ExpressionValueConverter(AbstractParticleValueConverter):
    def matches(self, value: str) -> bool:
        return value.startswith("exp(") and value.endswith(")")

    def convert(self, value: str) -> Any:
        exp = value[value.find("(") + 1 : value.find(")")]
        return SafelyExpEvalute(exp, eval_locals=VCONVERTER_CONSTS).safe_evalute()


PARTICLE_CONVERTERS: frozenset[AbstractParticleConverter] = frozenset(
    (GetattributeParticleConverter(),)
)
GETATTRIBUTE_VALUE_CONVERTERS: frozenset[AbstractParticleValueConverter] = frozenset(
    (
        BooleanValueConverter(),
        DigitValueConverter(),
        ListValueConverter(),
        ExpressionValueConverter(),
    )
)



# Copyright 2022 Animatea
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Adapters of aiotense.application.ports."""
from __future__ import annotations

__all__ = ["DigitParser", "TimedeltaParser"]

from typing import TYPE_CHECKING, Callable, Optional

from aiotense.application.ports import parsers

from . import converters

if TYPE_CHECKING:
    from aiotense.application.ports import converters as abc_converters
    from aiotense.domain import model


class DigitParser(parsers.AbstractParser):
    async def _parse(self, raw_str: str) -> int:
        multiplier = self.tense.multiplier
        resolved = self._resolver(raw_str, self.tense)
        duration = 0

        for pos, word in enumerate(resolved):
            for unit in self.tense:
                if word in unit.aliases:
                    prev_entry = resolved[pos - 1]
                    if not prev_entry.isdigit():
                        continue

                    duration += int(prev_entry) * (unit.duration * multiplier)

        return duration


class TimedeltaParser(DigitParser):
    def __init__(
        self,
        *,
        tense: model.Tense,
        resolver: Optional[Callable[[str, model.Tense], list[str]]] = None,
        converter: Optional[abc_converters.AbstractConverter] = None,
    ) -> None:
        super().__init__(
            tense=tense,
            resolver=resolver,
            converter=converters.TimedeltaConverter(),
        )



# Copyright 2022 Animatea
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Adapters of aiotense.application.ports."""
from __future__ import annotations

__all__ = ["TenseRepository"]

import copy
import dataclasses
from typing import Any, Optional, Type

from aiotense.application.ports.repository import AbstractTenseRepository
from aiotense.domain import units


class TenseRepository(AbstractTenseRepository):
    # <inherited docstring from :class:`TenseRepository`> #
    _tense_config: dict[str, Any] = {
        "model.Tense": {
            "multiplier": 1,
            "virtual": [],
        },
        "units.Second": {
            "duration": 1,
            "aliases": [
                "s",
                "sec",
                "secs",
                "second",
                "seconds",
            ],
        },
        "units.Minute": {
            "duration": 60,
            "aliases": [
                "m",
                "min",
                "mins",
                "minute",
                "minutes",
            ],
        },
        "units.Hour": {
            "duration": 60 * 60,
            "aliases": [
                "h",
                "hour",
                "hours",
            ],
        },
        "units.Day": {
            "duration": 60 * 60 * 24,
            "aliases": [
                "d",
                "day",
                "days",
            ],
        },
        "units.Week": {
            "duration": 60 * 60 * 24 * 7,
            "aliases": [
                "w",
                "week",
                "weeks",
            ],
        },
        "units.Year": {
            "duration": 60 * 60 * 24 * 365,
            "aliases": [
                "y",
                "year",
                "years",
            ],
        },
    }

    def __init__(self, config: Optional[dict[str, Any]] = None, /) -> None:
        if config is None:
            config = self._tense_config
        super().__init__(config)

    def get_config(self) -> dict[str, Any]:
        return copy.deepcopy(self._config)

    def get_setting(self, path: str, setting: str, /) -> Any:
        return self._config[path][setting]

    def add_virtual_unit(self, unit: units.VirtualUnit) -> TenseRepository:
        # <inherited docstring from :class:`TenseRepository`> #
        self._config["model.Tense"]["virtual"].append(dataclasses.asdict(unit))
        return self

    def add_virtual_unit_dict(self, unit_dict: dict[str, Any]) -> TenseRepository:
        # <inherited docstring from :class:`TenseRepository`> #
        self._config["model.Tense"]["virtual"].append(unit_dict)
        return self

    def add_aliases_to(
        self, unit: Type[units.Unit], aliases: list[str]
    ) -> TenseRepository:
        # <inherited docstring from :class:`TenseRepository`> #
        self._config["units." + unit.__name__]["aliases"].extend(aliases)
        return self



# Copyright 2022 Animatea
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tense domain."""
from __future__ import annotations

__all__ = ["Tense"]

import warnings
from dataclasses import dataclass, field
from typing import Any, Iterator

from aiotense.domain import units


@dataclass
class Tense:
    second: units.Second
    minute: units.Minute
    hour: units.Hour
    day: units.Day
    week: units.Week
    year: units.Year
    multiplier: int = 1
    virtual: list[dict[str, Any]] = field(default_factory=list)

    def __post_init__(self) -> None:
        self._cached_units: list[units.Unit] = []
        if self.multiplier <= 0:
            warnings.warn(
                "The time multiplier is less than zero, the work of "
                "parsers may be incorrect. It is recommended to set "
                "the value more than zero."
            )
        if self.virtual:
            self._resolve_virtual()

    def __iter__(self) -> Iterator[units.Unit]:
        _cached_units = self._cached_units
        if _cached_units:
            yield from _cached_units
        else:
            for unit in self.__dict__.values():
                if isinstance(unit, units.Unit):
                    self._cached_units.append(unit)
                    yield unit

    def _resolve_virtual(self) -> None:
        for n, unit_dict in enumerate(self.virtual):
            self.__dict__[f"virtual{n}"] = units.VirtualUnit(**unit_dict)

    @classmethod
    def from_dict(cls, tense_dict: dict[str, Any], /) -> Tense:
        tense_attrs = {}
        for key, attrs in tense_dict.items():
            module, cls_name = key.split(".")
            if cls_name == cls.__name__:
                if not isinstance(attrs, dict):
                    continue

                tense_attrs.update(attrs)
                continue

            module = globals()[module]
            unit_cls = getattr(module, cls_name)
            tense_attrs[cls_name.lower()] = unit_cls(**attrs)
            continue

        return cls(**tense_attrs)



# Copyright 2022 Animatea
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Dot_tense service domain."""
__all__ = ["HashableParticle", "HeaderParticle", "GetattributeParticle"]

import abc
from typing import cast


class HashableParticle(abc.ABC):
    """Base particle (token) class.
    Used at the step of lexical analysis for convenience when parsing.
    """

    def __init__(self, target: str) -> None:
        self.target = target

    def __hash__(self) -> int:
        return hash(self.target)

    @abc.abstractmethod
    def matches(self) -> bool:
        """Returns True if particle matches to current value (target)."""
        ...

    @classmethod
    def matches_converter(cls, converter: type | object) -> bool:
        if not hasattr(converter, "__name__"):
            converter = converter.__class__

        converter = cast(type, converter)
        return cls.__name__ in converter.__name__


class HeaderParticle(HashableParticle):
    """Header particle (token).

    Matches to:
     * [...]
    """

    def matches(self) -> bool:
        # <inherited docstring from :class:`HashableParticle` #
        return self.target.startswith("[") and self.target.endswith("]")

    @staticmethod
    def strip_header(header: str) -> str:
        return header[1:-1]


class GetattributeParticle(HashableParticle):
    """Getattribute particle (token).

    Matches to:
     * variable = ...
    """

    def matches(self) -> bool:
        # <inherited docstring from :class:`HashableParticle` #
        return len(self.target.split("=")) > 1



# Copyright 2022 Animatea
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Safely expression execution tool."""
__all__ = ["SafelyExpEvalute"]

import copy
import sys
import textwrap
import traceback
import types
from dataclasses import dataclass, field
from typing import Any, Final

_TAB: Final[str] = " " * 4


def _base_exc_format(exc: BaseException) -> str:
    return "\n".join(
        traceback.format_exception(
            type(exc),
            exc,
            exc.__traceback__,
        )
    )


@dataclass
class SafelyExpEvalute:
    """Class that represents safely expression evalute tool.
    This eval is safe because it creates a new dictionary of
    local variables for executing code that copied from the
    frame's local variables, which prevents global values from
    being manipulated.

    Parameters:
    -----------
    exp: :class:`str`
        Expression string.

    eval_locals: :class:`dict[str, Any]` = dict()
        Locals for expression execution.

    frame: :class:`types.FrameType` = sys._getframe(0)
        Frame for safely evalute.

    Examples:
    ---------
    # Base operations.
    >>> exp = "2 + 2 * 2"
    >>> assert SafelyExpEvalute(exp).safe_evalute() == 6

    # Flexible.
    >>> deps = {"minute": 60}
    >>> exp = "minute * 60"
    >>> assert SafelyExpEvalute(exp, eval_locals=deps).safe_evalute() == 3600
    """

    exp: str
    eval_locals: dict[str, Any] = field(default_factory=dict)
    frame: types.FrameType = field(default=sys._getframe(0))

    def __post_init__(self) -> None:
        self._add_expression_to_func()

    def _evaldict(self) -> dict[str, Any]:
        frame = self.frame
        safely_dict = copy.copy(frame.f_globals)
        safely_dict.update(frame.f_locals)
        return safely_dict

    def _add_expression_to_func(self) -> None:
        deps = {
            "func_name": self.func_name,
            "exp": textwrap.indent("return" + " " + self.exp, _TAB),
        }
        self._exp = "def {func_name}():\n{exp}".format(**deps)

    def safe_evalute(self) -> Any:
        evaldict = self._evaldict()
        try:
            code = compile(self._exp, self.frame.f_code.co_filename, "single")
            if self.eval_locals:
                evaldict.update(self.eval_locals)

            exec(code, evaldict)
        except BaseException as exc:
            return _base_exc_format(exc)

        func = evaldict[self.func_name]
        return func()

    @property
    def func_name(self) -> str:
        return "__" + self.__class__.__name__.lower()



# Copyright 2022 Animatea
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Functional project tools."""
__all__ = ["async_lru"]

import functools
from collections import OrderedDict
from typing import Any, Awaitable, Callable, Coroutine, ParamSpec, TypeVar, cast

P = ParamSpec("P")
RT = TypeVar("RT")


def async_lru(
    maxsize: int = 128,
) -> Callable[[Callable[P, Awaitable[RT]]], Callable[P, Awaitable[RT]]]:
    """LRU cache wrapper for coroutines.
    A simple LRU cache that based on collections.OrderedDict where the key is the
        ~ function name + (positional arguments) + {keyword arguments}

    Parameters:
    -----------
    maxsize: :class:`int` = 128
        Cache size.
    """
    cache: OrderedDict[str, Any] = OrderedDict()

    def inner(
        fn: Callable[P, Awaitable[RT]],
    ) -> Callable[P, Coroutine[Any, Any, RT]]:
        @functools.wraps(fn)
        async def wrapper(*args: P.args, **kwargs: P.kwargs) -> RT:
            key = getattr(fn, "__name__") + str((args, kwargs))
            if key not in cache:
                if len(cache) >= maxsize:
                    cache.popitem(last=False)

                cache[key] = await fn(*args, **kwargs)
            return cast(RT, cache[key])

        return wrapper

    return inner



# Copyright 2022 Animatea
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Interfaces that are implemented in aiotense.adapters."""
from __future__ import annotations

__all__ = ["AbstractParser"]

import abc
from typing import TYPE_CHECKING, Any, Callable, Optional, final

from aiotense.application.ports import converters

if TYPE_CHECKING:
    from aiotense.domain import model


class AbstractParser(abc.ABC):
    def __init__(
        self,
        *,
        tense: model.Tense,
        resolver: Optional[Callable[[str, model.Tense], list[str]]] = None,
        converter: Optional[converters.AbstractConverter] = None,
    ) -> None:
        self.tense = tense
        self.converter = converter
        self._resolver = resolver

    @final
    async def parse(self, raw_str: str) -> Any:
        value = await self._parse(raw_str)
        if self.converter is not None:
            value = await self.converter.convert(value)

        return value

    @abc.abstractmethod
    def _parse(self, raw_str: str) -> Any:
        ...

    @property
    def resolver(self) -> Optional[Callable[[str, model.Tense], list[str]]]:
        return self._resolver

    @resolver.setter
    def resolver(self, new_resolver: Callable[[str, model.Tense], list[str]]) -> None:
        if not callable(new_resolver):
            raise ValueError("Resolver must be callable.")
        self._resolver = new_resolver



# Copyright 2022 Animatea
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tense domain."""
from __future__ import annotations

__all__ = ["Unit", "Minute", "Hour", "Day", "Week", "VirtualUnit", "Second"]

import warnings
from dataclasses import dataclass


@dataclass
class Unit:
    aliases: list[str]
    duration: int

    def __post_init__(self) -> None:
        if self.duration <= 0:
            warnings.warn(
                "The unit duration is less than zero, the work of "
                "parsers may be incorrect. It is recommended to set "
                "the value more than zero."
            )


@dataclass
class VirtualUnit(Unit):
    """Used to create custom units of time."""

    pass


@dataclass
class Second(Unit):
    duration: int = 1


@dataclass
class Minute(Unit):
    duration: int = 60


@dataclass
class Hour(Unit):
    duration: int = 60 * 60


@dataclass
class Day(Unit):
    duration: int = 60 * 60 * 24


@dataclass
class Week(Unit):
    duration: int = 60 * 60 * 24 * 7


@dataclass
class Year(Unit):
    duration: int = 60 * 60 * 24 * 365



# Copyright 2022 Animatea
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Time resolvers."""
from __future__ import annotations

__all__ = ["basic_resolver", "smart_resolver"]

import re
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from aiotense.domain import model

DIGIT_PATTERN: re.Pattern[str] = re.compile(r"(\d+)")


def basic_resolver(raw_str: str, _: model.Tense) -> list[str]:
    """Example of supported patterns:
    * '1d1min'
    * '1d 1min'
    * '1d1min 2 seconds'
    * etc
    """
    return list(
        filter(
            bool,
            DIGIT_PATTERN.split(raw_str.replace(" ", "")),
        )
    )


def smart_resolver(raw_str: str, tense: model.Tense) -> list[str]:
    """Example of supported patterns:
    * All patterns from `resolve_time_string`
    * 1year and 10 months + 5 seconds
    * etc
    """
    basic_resolve = basic_resolver(raw_str, tense)
    if any(not p.isalpha() for p in basic_resolve if not p.isdigit()):
        # Removes any char of string.punctuation
        def _resolve_p(p: str, /) -> str:
            return p if p.isdigit() else "".join(filter(str.isalpha, p))

        basic_resolve = [_resolve_p(p) for p in basic_resolve[:]]

    unit_aliases: list[str] = sum((u.aliases for u in tense), [])
    for idx, part in enumerate(basic_resolve):
        if part.isdigit():
            continue

        part = part.lower()  # case insensitive
        for alias in unit_aliases:
            if alias in part:
                if len(alias) == 1 and len(part) > 1:
                    # For 'd', 's', 'y', ... aliases.
                    continue
                basic_resolve.pop(idx)
                basic_resolve.insert(idx, alias)

    return basic_resolve



# Copyright 2022 Animatea
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Unit of work for atomic alias operations."""
from __future__ import annotations

__all__ = ["AbstractTenseUnitOfWork", "TenseUnitOfWork"]

import abc
from typing import TYPE_CHECKING, Any, Iterable, Type

from aiotense.adapters import repository
from aiotense.domain import units

if TYPE_CHECKING:
    from aiotense.application.ports import repository as abc_repository


class AbstractUnitOfWork:
    """Base class for atomic operations. Useful for annotations.

    Implements in:
        ~ :class:`AbstractTenseUnitOfWork`
        ~ :class:`TenseUnitOfWork`
    """

    products: Any
    pass


class AbstractTenseUnitOfWork(AbstractUnitOfWork, abc.ABC):
    """An abstract class whose subclasses are responsible for
    atomic operations on Tense units (remove unit aliases/replace/...).

    Implements in:
        ~ :class:`TenseUnitOfWork`
    """

    tenses: abc_repository.AbstractTenseRepository

    def __enter__(self) -> AbstractTenseUnitOfWork:
        return self

    def __exit__(self, *args: Any) -> None:
        """Stub for syntax sugar `with`. Allows you to group code into
        logical blocks.
        """
        pass

    @staticmethod
    def with_unit_resolve(unit: str | Type[units.Unit], /) -> str:
        if not isinstance(unit, str):
            if not issubclass(unit, units.Unit):
                raise ValueError("Unit must be instance of str or units.Unit.")
            unit = unit.__name__
        if not unit.startswith("units."):
            unit = "units." + unit.title()
        return unit

    @abc.abstractmethod
    def update_config(self, config: dict[str, Any], /) -> None:
        """Replaces the items of the current dictionary config with new
        ones using the dict.update(new) method.

        Parameters:
        -----------
        config: :class:`dict[str, Any]`, /
            New config items.

        Returns:
        --------
        builtins.None
        """
        ...

    @abc.abstractmethod
    def delete_aliases(self, unit: Type[units.Unit], aliases: Iterable[str]) -> None:
        """Removes aliases for a concrete unit of time.

        !!! Note:
            Will raise KeyError if you specify a non-existent unit of time.
            See the docs for all valid time units. If you added
            virtual aliases, then you can view the current state of the
            config using the `~self.get_config()` method or by accessing
            the config attribute (`~self.products.config`).

        Parameters:
        -----------
        unit: :class:`str`
            Concrete unit of time.

        aliases: :class:`Iterable[str]`
            Iterable object that contains string aliases for concrete unit
            of time.

        Returns:
        --------
        builtins.None
        """
        ...

    @abc.abstractmethod
    def replace_aliases(
        self, unit: Type[units.Unit], replacements: dict[str, str]
    ) -> None:
        """Replace aliases for concrete unit of time.

        !!! Note:
            Will raise KeyError if you specify a non-existent unit of time.
            See the docs for all valid time units. If you added
            virtual aliases, then you can view the current state of the
            config using the `~self.get_config()` method or by accessing
            the config attribute (`~self.products.config`).

        Parameters:
        -----------
        unit: :class:`str`
            Concrete unit of time.

        replacements: :class:`dict[str, str]`
            Dictionary where keys are old aliases and values are new aliases.

        Returns:
        --------
        builtins.None
        """
        ...


class TenseUnitOfWork(AbstractTenseUnitOfWork):
    """Adapter of :class:`AbstractTenseUnitOfWork`."""

    def __enter__(self) -> AbstractTenseUnitOfWork:
        self.tenses = repository.TenseRepository()
        return super().__enter__()

    def update_config(self, config: dict[str, Any]) -> None:
        # <<inherited docstring from :class:`AbstractTenseUnitOfWork`>> #
        self.tenses._config.update(config)

    def delete_aliases(self, unit: Type[units.Unit], aliases: Iterable[str]) -> None:
        # <<inherited docstring from :class:`AbstractTenseUnitOfWork`>> #
        unit = self.with_unit_resolve(unit.__name__)  # type: ignore[assignment]
        unit_aliases = self.tenses._config[unit]["aliases"]
        for alias in aliases:
            unit_aliases.remove(alias)

    def replace_aliases(
        self, unit: Type[units.Unit], replacements: dict[str, str]
    ) -> None:
        # <<inherited docstring from :class:`AbstractTenseUnitOfWork`>> #
        unit = self.with_unit_resolve(unit.__name__)  # type: ignore[assignment]
        unit_aliases = self.tenses._config[unit]["aliases"]
        for old, new in replacements.items():
            for idx, exiting_alias in enumerate(unit_aliases):
                if old == exiting_alias:
                    unit_aliases.pop(idx)
                    unit_aliases.insert(idx, new)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['aiotense',
 'aiotense.adapters',
 'aiotense.application',
 'aiotense.application.ports',
 'aiotense.domain',
 'aiotense.service_layer',
 'aiotense.service_layer.dot_tense']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'aiotense',
    'version': '0.0.2a0',
    'description': 'Time Processing Tool',
    'long_description': '<div id="top"></div>\n<img src="assets/tense-logo.jpg" align="left" width="200px"/>\nProject: aiotense\n<br>\nLicense: Apache 2.0\n<br>\nAbout: Time Processing Tool\n<br>\nOS: Independent\n<br>\nPython: 3.9+\n<br>\nTyping: Typed\n<br>\nTopic: Utilities\n<br />\n    <p align="center">\n    <br />\n    <a href="https://animatea.github.io/aiotense/">Documentation</a>\n    ·\n    <a href="https://github.com/Animatea/aiotense/issues">Report Bug</a>\n    ·\n    <a href="https://github.com/Animatea/aiotense/issues">Request Feature</a>\n    </p>\n<div id="top"></div>\n<p align="center">\n   <a href="i18n/ua_README.md"><img height="20" src="https://img.shields.io/badge/language-ua-green?style=social&logo=googletranslate"></a>\n   <a href="i18n/ru_README.md"><img height="20" src="https://img.shields.io/badge/language-ru-green?style=social&logo=googletranslate"></a>\n</p>\n<details>\n  <summary>Table of Contents</summary>\n  <ol>\n    <li>\n      <a href="#about-the-project">About The Project</a>\n      <ul>\n        <li><a href="#welcome">Welcome</a></li>\n      </ul>\n    </li>\n    <li>\n      <a href="#getting-started">Getting Started</a>\n      <ul>\n        <li><a href="#with-pypi">With PyPi</a></li>\n        <li><a href="#with-poetry">With Poetry</a></li>\n      </ul>\n    </li>\n    <li>\n      <a href="#usage">Usage</a>\n      <ul>\n        <li><a href="#built-in-basic">Built-in basics</a></li>\n        <li><a href="#reconfiguring-existing-settings">Reconfiguring existing settings</a></li>\n        <li><a href="#adding-new-settings">Adding new settings</a></li>\n        <li><a href="#faq">FAQ</a></li>\n      </ul>\n    </li>\n    <li><a href="#examples">Examples</a></li>\n    <li><a href="#contributing">Contributing</a></li>\n    <li><a href="#license">License</a></li>\n    <li><a href="#contact">Contact</a></li>\n    <li><a href="#acknowledgments">Acknowledgments</a></li>\n  </ol>\n</details>\n\n## About The Project\n<a href="https://circleci.com/gh/Animatea/aiotense/tree/main"><img height="20" src="https://circleci.com/gh/Animatea/aiotense/tree/main.svg?style=svg&circle-token=066eb74fc70db6eeaa6df5ade9e3d2df131b1de1"></a>\n<a href="https://pypi.org/project/python-multibar/"><img height="20" alt="Pypi" src="https://img.shields.io/pypi/v/aiotense"></a>\n<a href="https://pypi.org/project/mypy/"><img height="20" alt="Mypy badge" src="http://www.mypy-lang.org/static/mypy_badge.svg"></a>\n<a href="https://github.com/psf/black"><img height="20" alt="Black" src="https://img.shields.io/badge/code%20style-black-000000.svg"></a>\n<a href="https://pycqa.github.io/isort/"><img height="20" alt="Supported python versions" src="https://img.shields.io/badge/%20imports-isort-%231674b1?style=flat&labelColor=ef8336">\n<a href="https://pypi.org/project/python-multibar/"><img height="20" alt="Versions" src="https://img.shields.io/pypi/pyversions/aiotense"></a>\n\n### Welcome\n> Have you ever needed to convert, for example, the string "1d1minute 2 sec" \nto the number of seconds or a datetime.timedelta object?\n\nNo? Then advise us to your friends :) And if you really need our tool - let\'s move on!\n<h5 align="center">UML Diagram with the main logic of the project</h5>\n<img src="assets/tense-uml.jpg" align="left"/>\n\n<p align="right"><a href="#top"><img height="20" src="https://img.shields.io/badge/back_to-top-green?style=social&logo=github"></a></p>\n\n## Getting started\n### With PyPi\n```bash\n$ pip3 install aiotense\n```\n\n### With Poetry\n```bash\n$ poetry add aiotense\n```\n<p align="right"><a href="#top"><img height="20" src="https://img.shields.io/badge/back_to-top-green?style=social&logo=github"></a></p>\n\n## Usage\n### Built-in basic\n```py\nimport asyncio\nimport datetime\nfrom aiotense import TenseParser\n\ntime_string = "1d2minutes 5 sec"\n\n# <-- Digit parser -->\ndigit_parser = TenseParser(TenseParser.DIGIT)\ndigit_value = asyncio.run(digit_parser.parse(time_string))\n# <-- Assertions -->\nassert digit_value == 86525\n\n# <-- Timedelta parser -->\ndelta_parser = TenseParser(TenseParser.TIMEDELTA)\ndelta_value = asyncio.run(delta_parser.parse(time_string))\n# <-- Assertions -->\nassert isinstance(delta_value, datetime.timedelta)\nassert str(delta_value) == "1 day, 0:02:05"\n```\n<p align="right"><a href="#top"><img height="20" src="https://img.shields.io/badge/back_to-top-green?style=social&logo=github"></a></p>\n\n### Reconfiguring existing settings\n```py\nimport asyncio\nfrom aiotense import TenseParser, from_tense_file_source\n\nconfig_emulation = """\n[model.Tense]\nmultiplier = 2  # each unit of time will be multiplied by 2\n# !!! Note: If the multiplier is <= 0, then the parsers will \n# not work correctly. In this case, a warning will be sent to the console.\n\n[units.Minute]\nduration = 120  # Why not?...\naliases = my_minute, my_minutes, my_min, my_mins\n"""\nparser = TenseParser(\n    TenseParser.TIMEDELTA,\n    config=from_tense_file_source(config_emulation),\n)\ndelta_value = asyncio.run(parser.parse("1 my_min 10my_mins 9  my_minutes"))\n# <-- Assertions -->\nassert str(delta_value) ==  "1:20:00" # (each 120 * 2)\n```\n<p align="right"><a href="#top"><img height="20" src="https://img.shields.io/badge/back_to-top-green?style=social&logo=github"></a></p>\n\n### Adding new settings\n```py\nimport asyncio\nfrom aiotense import TenseParser, from_tense_file_source\n\nconfig_emulation = """\n[model.Tense]  # This header is required.\n\n[virtual]\nduration = exp(year * 10)\naliases = decade, dec, decs, decades\n"""\n\nparser = TenseParser(\n    TenseParser.TIMEDELTA,\n    config=from_tense_file_source(config_emulation),\n)\ndelta_value = asyncio.run(parser.parse("1year 10 decades5   seconds"))\n# <-- Assertions -->\nassert str(delta_value) == "36865 days, 0:00:05"\n```\n<p align="right"><a href="#top"><img height="20" src="https://img.shields.io/badge/back_to-top-green?style=social&logo=github"></a></p>\n\n### FAQ\nBut what if you need to parse a string like: "1day and 10 minutes + 5 seconds"?\nLet\'s see:\n```py\n>>> import asyncio\n>>> from aiotense import TenseParser\n\n>>> complex_string = "1day and 10 minutes + 5 seconds"\n\n>>> parser = TenseParser(TenseParser.TIMEDELTA)\n>>> asyncio.run(parser.parse(complex_string))\n\'0:00:05\'\n```\nWait... What? 5 second? But there are days and minutes...\n- It\'s okay, you\'re using flexible aiotense! This problem is solved in two ways:\n  1) You write your own time_resolver and pass it\n  2) Choose an existing one from aiotense.resolvers\n\nLet\'s demonstrate!\nI will use the second option, since the built-in time resolvers in aiotense are suitable for me.\n```py\n>>> import asyncio\n>>> from aiotense import TenseParser, resolvers\n\n>>> complex_string = "1day and 10 minutes + 5 seconds"\n\n>>> parser = TenseParser(TenseParser.TIMEDELTA, time_resolver=resolvers.smart_resolver)\n>>> asyncio.run(parser.parse(complex_string)) \n\'1 day, 0:10:05\'\n```\nWell, that\'s better!\n\n**aiotense.application.resolvers.smart_resolver()** is also case insensitive!\n```py\n>>> import asyncio\n>>> from aiotense import TenseParser, resolvers\n\n>>> complex_string = "1DAY and 10 MINUTES + 5 SECONDS"\n\n>>> parser = TenseParser(TenseParser.TIMEDELTA, time_resolver=resolvers.smart_resolver)\n>>> asyncio.run(parser.parse(complex_string)) \n\'1 day, 0:10:05\'\n```\n\n<p align="right"><a href="#top"><img height="20" src="https://img.shields.io/badge/back_to-top-green?style=social&logo=github"></a></p>\n\n## Examples.\nIf you think that this is where the possibilities of aiotense ends, then you are wrong! \nThe possibilities of aiotense are too many for a README, so I suggest you move on to viewing \nthe usage examples here:\n<p align="center">\n<br />\n<a href="https://github.com/Animatea/aiotense/tree/main/examples">Aiotense Examples</a>\n</p>\n<p align="right"><a href="#top"><img height="20" src="https://img.shields.io/badge/back_to-top-green?style=social&logo=github"></a></p>\n\n## Contributing\n\nContributions are what make the open source community such an amazing place to learn, inspire, and create. Any contributions you make are **greatly appreciated**.\n\nIf you have a suggestion that would make this better, please fork the repo and create a pull request. You can also simply open an issue with the tag "enhancement".\nDon\'t forget to give the project a star! Thanks again!\n\n1. Fork the Project\n2. Create your Feature Branch (`git checkout -b feature/AmazingFeature`)\n3. Commit your Changes (`git commit -m \'Add some AmazingFeature\'`)\n4. Push to the Branch (`git push origin feature/AmazingFeature`)\n5. Open a Pull Request\n<p align="right"><a href="#top"><img height="20" src="https://img.shields.io/badge/back_to-top-green?style=social&logo=github"></a></p>\n\n<!-- LICENSE -->\n## License\n\nDistributed under the Apache 2.0 License. See [`LICENSE`](https://github.com/Animatea/aiotense/blob/main/LICENSE) for more information.\n\n<p align="right"><a href="#top"><img height="20" src="https://img.shields.io/badge/back_to-top-green?style=social&logo=github"></a></p>\n\n\n<!-- CONTACT -->\n## Contact\n<div align="left">\n    <a href="https://discord.com/invite/KKUFRZCt4f"><img src="https://discordapp.com/api/guilds/744099317836677161/widget.png?style=banner4" alt="" /></a>\n</div>\n\n<p align="right"><a href="#top"><img height="20" src="https://img.shields.io/badge/back_to-top-green?style=social&logo=github"></a></p>\n\n\n## Acknowledgments\n* [Choose an Open Source License](https://choosealicense.com)\n* [Img Shields](https://shields.io)\n* [GitHub Pages](https://pages.github.com)\n* [Python](https://www.python.org)\n* [Python Community](https://www.python.org/community/)\n* [MkDocs](https://www.mkdocs.org)\n* [MkDocs Material](https://squidfunk.github.io/mkdocs-material/)\n\n<p align="right"><a href="#top"><img height="20" src="https://img.shields.io/badge/back_to-top-green?style=social&logo=github"></a></p>\n',
    'author': 'Animatea',
    'author_email': 'animatea.programming@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/Animatea/aiotense',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



# Copyright 2022 Animatea
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Adapters of aiotense.application.ports."""
__all__ = ["TimedeltaConverter"]

from datetime import timedelta
from typing import Any

from aiotense.application.ports import converters


class TimedeltaConverter(converters.AbstractConverter[timedelta]):
    async def convert(self, value: Any) -> timedelta:
        if not isinstance(value, (int, float)):
            raise ValueError("Value must be instance of (int, float).")

        return timedelta(seconds=float(value))



# Copyright 2022 Animatea
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Aiotense factory."""
from __future__ import annotations

__all__ = ["TenseParser"]

from typing import TYPE_CHECKING, Any, Callable, Optional, Type

from aiotense.adapters import parsers, repository
from aiotense.domain import model

if TYPE_CHECKING:
    from aiotense.application.ports import converters

from . import exceptions, resolvers
from .ports import parsers as abc_parsers

_tenses = repository.TenseRepository()


class TenseParser:
    """Base parsers factory.

    Parameters:
    -----------
    parser_cls: :class:`Any` = DIGIT
        Concrete parser type.
    tense: :class:`model.Tense` = model.Tense.from_dict(_tenses.config), *
        Configuration for concrete parser.

    Raises:
    -------
    :class:`exceptions.InvalidParserType`
        Raises if 'parser_cls' is not subclass of :class:`abc_parsers.AbstractParser`
    """

    TIMEDELTA = parsers.TimedeltaParser
    DIGIT = parsers.DigitParser

    def __new__(
        cls,
        parser_cls: Type[abc_parsers.AbstractParser] = DIGIT,
        *,
        config: Optional[dict[str, Any]] = None,
        converter: Optional[converters.AbstractConverter] = None,
        time_resolver: Optional[Callable[[str, model.Tense], list[str]]] = None,
    ) -> abc_parsers.AbstractParser:
        if config is not None:
            _tenses._config.update(config)

        if time_resolver is None:
            time_resolver = resolvers.basic_resolver

        if not issubclass(parser_cls, abc_parsers.AbstractParser):
            raise exceptions.InvalidParserType(
                f"Invalid parser type, you can only use {parsers.__all__}."
            )
        instance = parser_cls.__new__(parser_cls)  # type: ignore[call-overload]
        instance.__init__(
            tense=model.Tense.from_dict(_tenses.config),
            resolver=time_resolver,
            converter=converter,
        )
        return instance



# Copyright 2022 Animatea
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tense config parser."""
from __future__ import annotations

__all__ = ["from_tense_file", "from_tense_file_source"]

import abc
import difflib
import inspect
import pathlib
from typing import (
    Any,
    Final,
    Generic,
    Hashable,
    Iterable,
    Optional,
    Type,
    TypeVar,
    cast,
    final,
)

from aiotense.domain import model, units
from aiotense.service_layer.dot_tense import converters, domain, exceptions

T = TypeVar("T")
T_co = TypeVar("T_co", covariant=True)

_COMMENT: Final[str] = "#"
_BASE_SETTING_PATH: Final[str] = "model.Tense"
_VIRTUAL_PREFIX: Final[str] = "virtual"
_VIRTUALS: list[dict[str, Any]] = []

_ONEWORD_HEADER_MATCHES: Final[tuple[str, ...]] = (_VIRTUAL_PREFIX,)
_MODULE_HEADER_MATCHES: Final[tuple[str, ...]] = (
    "model",
    "units",
)
_OBJTYPE_MATCHES: Final[list[str]] = model.__all__ + units.__all__


def sorting_hat(initial: str, /) -> domain.HashableParticle:
    """Converts a string to a particle (token) for further parsing.
    Checks all domain names, if the class is abstract then skips it.

    Parameters
    -----------
    initial: :class:`str`, /
        String to parse.

    Raises
    ------
    :class:`KeyError`
        If particle for string is not found.
    """
    for pname in domain.__all__:
        particle = getattr(domain, pname)
        if inspect.isabstract(particle):
            continue

        particle = particle(initial)
        if particle.matches():
            return cast(domain.HashableParticle, particle)

    raise KeyError(f"Particle not found for {initial!s}.")


def inject_particle_converters() -> dict[
    Type[domain.HashableParticle],
    converters.AbstractParticleConverter,
]:
    """Injects converters for particles.
    Abstract particles are skipped.
    """
    injected = {}
    for pname in domain.__all__:
        particle = getattr(domain, pname)
        if inspect.isabstract(particle):
            continue

        for converter in converters.PARTICLE_CONVERTERS:
            if particle.matches_converter(converter):
                injected[particle] = converter
                continue

    return injected


class AbstractStepChain(abc.ABC, Generic[T, T_co]):
    """Abstract class - chain of responsibility pattern.
    Used to step by step parse the configuration file into a working state.

    !!! Note:
        This class is Generic in which:
        ~T - accept value.
        ~T_co - return value.
    """

    def __init__(self) -> None:
        self._next_handler: Optional[AbstractStepChain[T, T_co]] = None

    @final
    def set_next(
        self, handler: AbstractStepChain[Any, Any]
    ) -> AbstractStepChain[Any, Any]:
        self._next_handler = handler
        return handler

    @abc.abstractmethod
    def take_a_step(self, target: T) -> T_co:
        """Converts the current value - (~T) to another value - (~T_co)."""
        ...

    @property
    def next_handler(self) -> Optional[AbstractStepChain[Any, Any]]:
        return self._next_handler


class LexingStep(AbstractStepChain[str, dict[str, Any]]):
    """At this step, a string representation of the configuration is received.
    Breaks the string config into particles (tokens) for further parsing.

    !!! Note:
        * The parser goes through the configuration from top to bottom, one line
          at a time. Therefore, it is important to observe the order of settings.
        * Each header is a dictionary.
    """

    @staticmethod
    def _trim_comment(line: str, /) -> str:
        """Removes comments from a string line."""
        if _COMMENT in line:
            return line[: line.find(_COMMENT)]
        return line

    def take_a_step(self, target: str) -> dict[str, Any]:
        # <inherited docstring from :class:`AbstractStepChain`> #
        groups: dict[str, Any] = {}
        last_section = ""
        pconverters = inject_particle_converters()

        for line in target.split("\n"):
            line = self._trim_comment(line).strip()
            if not line:
                # Line is empty.
                continue

            particle = sorting_hat(line)  # Parsing line to particle (token).
            particle_type = type(particle)
            if isinstance(particle, domain.HeaderParticle):
                # Currently, supports only comma lists
                # | alias, alias, alias
                section = particle.strip_header(line)
                groups[section] = {}
                last_section = section

            elif isinstance(particle, domain.GetattributeParticle):
                # Currently, supports only with exp() wrapper.
                # | exp(2 + 2 * 2)
                groups[last_section].update(
                    pconverters[particle_type].convert(particle.target)
                )

            continue

        return groups


class AnalyzeStep(AbstractStepChain[dict[str, Any], dict[str, Any]]):
    """At this stage, the already converted configuration into a dictionary
    is accepted. Checks for validity all names and in case of failure gives
    an error (often with a suggestion of possible variants of the word).
    """

    @staticmethod
    def _with_suggest_to(
        part: str,
        *,
        matches: Iterable[str],
        in_: Optional[str] = None,
    ) -> str:
        """Suggest possible word options."""
        base_msg = f"Undefined part {part!r}"
        if in_ is not None:
            base_msg += f" in {in_!r}."

        matches = difflib.get_close_matches(part, matches)
        if matches:
            base_msg += f" Did you mean any of ~{matches}?"
        return base_msg

    def take_a_step(self, target: dict[str, Any]) -> dict[str, Any]:
        # <inherited docstring from :class:`AbstractStepChain`> #
        _copied_target = target.copy()
        for key, value in _copied_target.items():
            key_parts = key.split(".")
            if len(key_parts) == 1:
                # In this case, the line is a one-word header, something like [virtual].
                # For such it is necessary to make a separate check.
                key = key_parts[0].lower()
                if key not in _ONEWORD_HEADER_MATCHES:
                    raise exceptions.AnalyzeError(
                        self._with_suggest_to(key, matches=_ONEWORD_HEADER_MATCHES)
                    )
                # Currently, supports only one one-word header.
                if key == _VIRTUAL_PREFIX:
                    _VIRTUALS.append(value)
                    target.pop(key)
                    continue

            key_type, obj_type = key_parts
            module = globals().get(key_type, None)
            if module is None:
                # Prefix like `model.Tense` or `units.Minute`.
                #              ^^^^^^           ^^^^^^
                raise exceptions.AnalyzeError(
                    self._with_suggest_to(
                        key_type, in_=key, matches=_MODULE_HEADER_MATCHES
                    )
                )

            obj = getattr(module, obj_type, None)
            if obj is None:
                # Object like `model.Tense` or `units.Minute`
                #                    ^^^^^            ^^^^^^
                raise exceptions.AnalyzeError(
                    self._with_suggest_to(obj_type, in_=key, matches=_OBJTYPE_MATCHES)
                )

        return target


class CompilingStep(AbstractStepChain[dict[str, Any], dict[str, Any]]):
    """At this step, virtual values are added to the already converted dictionary."""

    def take_a_step(self, target: dict[str, Any]) -> dict[str, Any]:
        # <inherited docstring from :class:`AbstractStepChain`> #
        if _VIRTUALS:
            target[_BASE_SETTING_PATH][_VIRTUAL_PREFIX] = _VIRTUALS
        return target


class _ShadowStep(AbstractStepChain[dict[Hashable, Any], dict[Hashable, Any]]):
    """The shadow step, which is necessary for the chain of responsibilities to work correctly."""

    def take_a_step(self, target: dict[Hashable, Any]) -> dict[Hashable, Any]:
        # <inherited docstring from :class:`AbstractStepChain`> #
        return target


def from_tense_file_source(file_source: str, /) -> Any:
    """Parses a configuration file into a dictionary.

    Parameters:
    -----------
    file_source: :class:`str`, /
        Config file string representation.
    """
    (
        (first_step := LexingStep())
        .set_next(AnalyzeStep())
        .set_next(CompilingStep())
        .set_next(_ShadowStep())
    )
    handler: AbstractStepChain[Any, Any] = first_step
    parsed_source: Any = file_source
    while handler.next_handler is not None:
        parsed_source = handler.take_a_step(parsed_source)
        handler = handler.next_handler

    return parsed_source


def from_tense_file(
    path: pathlib.Path | str,
    /,
    encoding: str = "utf-8",
) -> Any:
    """Opens the configuration file and parses source into a dictionary.

    Parameters:
    -----------
    path: :class:`Union[pathlib.Path, str]`, /
        Path to configuration file.
    encoding: :class:`Optional[str]` = None
        File encoding.
    """
    with open(path, "r", encoding=encoding) as file:
        return from_tense_file_source(file.read())



# Copyright 2022 Animatea
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Root package."""
from aiotense.application import TenseParser, resolvers
from aiotense.domain import model, units
from aiotense.service_layer import unit_of_work
from aiotense.service_layer.dot_tense import from_tense_file, from_tense_file_source

__all__ = [
    "model",
    "units",
    "resolvers",
    "TenseParser",
    "unit_of_work",
    "from_tense_file_source",
    "from_tense_file",
]



# Copyright 2022 Animatea
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Interfaces that are implemented in aiotense.adapters."""
from __future__ import annotations

__all__ = ["AbstractTenseRepository"]

import abc
from typing import TYPE_CHECKING, Any, Optional, Type, TypedDict, cast

if TYPE_CHECKING:
    from aiotense.domain import units

    class _VirtualUnitDict(TypedDict):
        duration: int
        aliases: list[str]


class AbstractTenseRepository(abc.ABC):
    """`Add` operations allows fluent-style."""

    def __init__(self, config: Optional[dict[str, Any]] = None, /) -> None:
        self._config = config

    def __getitem__(self, item: Any) -> Any:
        if not isinstance(item, str):
            return NotImplemented
        config = cast(dict[str, Any], self._config)
        return config[item]

    def __setitem__(self, key: Any, value: Any) -> Any:
        if not isinstance(key, str) or not isinstance(value, str):
            return NotImplemented
        if self._config is None:
            return NotImplemented
        if key not in self._config:
            raise KeyError(
                f"Key {key} does not exist in config. New keys cannot be added."
            )
        self._config[key] = value

    @property
    def config(self) -> Optional[dict[str, Any]]:
        return self._config

    @abc.abstractmethod
    def get_config(self) -> dict[str, Any]:
        """Returns deepcopy of current config."""
        ...

    @abc.abstractmethod
    def get_setting(self, path: str, setting: str, /) -> Any:
        ...

    @abc.abstractmethod
    def add_virtual_unit(self, unit: units.VirtualUnit) -> AbstractTenseRepository:
        """Adds custom unit of time. Allows fluent-style."""
        ...

    @abc.abstractmethod
    def add_virtual_unit_dict(
        self, unit_dict: _VirtualUnitDict
    ) -> AbstractTenseRepository:
        """Adds custom unit of time. Allows fluent-style."""
        ...

    @abc.abstractmethod
    def add_aliases_to(
        self, unit: Type[units.Unit], aliases: list[str]
    ) -> AbstractTenseRepository:
        """Adds aliases to unit of time. Allows fluent-style."""
        ...
